/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.server.mixin.entity.player;

import com.google.common.collect.Maps;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.util.BlockPos;
import net.minecraft.world.World;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Inject;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfo;
import org.spongepowered.server.interfaces.IMixinEntityPlayer;

import java.util.HashMap;
import java.util.Map;

@Mixin(value = EntityPlayer.class, priority = 1001)
public abstract class MixinEntityPlayer extends EntityLivingBase {
    private static final String PERSISTED_NBT_TAG = "PlayerPersisted";

    private HashMap<Integer, BlockPos> spawnChunkMap = Maps.newHashMap();
    private HashMap<Integer, Boolean> spawnForcedMap = Maps.newHashMap();

    @Shadow protected BlockPos spawnChunk;
    @Shadow private boolean spawnForced;

    protected MixinEntityPlayer(World worldIn) {
        super(worldIn);
    }

    @Inject(method = "setSpawnPoint", at = @At("HEAD"), cancellable = true)
    public void onSetSpawnPoint(BlockPos pos, boolean forced, CallbackInfo ci) {
        if (this.field_71093_bK != 0) {
            setSpawnChunk(pos, forced, this.field_71093_bK);
            ci.cancel();
        }
    }

    @Inject(method = "clonePlayer", at = @At("RETURN"))
    public void onClonePlayerEnd(EntityPlayer oldPlayer, boolean respawnFromEnd, CallbackInfo ci) {
        this.spawnChunkMap = ((IMixinEntityPlayer) oldPlayer).getSpawnChunkMap();
        this.spawnForcedMap = ((IMixinEntityPlayer) oldPlayer).getSpawnForcedMap();

        final NBTTagCompound old = ((IMixinEntityPlayer) oldPlayer).getEntityData();
        if (old.func_74764_b(PERSISTED_NBT_TAG)) {
            ((IMixinEntityPlayer) this).getEntityData().func_74782_a(PERSISTED_NBT_TAG, old.func_74775_l(PERSISTED_NBT_TAG));
        }
    }

    @Inject(method = "readEntityFromNBT", at = @At(value = "FIELD", target = "net.minecraft.entity.player.EntityPlayer"
            + ".foodStats:Lnet/minecraft/util/FoodStats;"))
    public void onReadEntityFromNBT(NBTTagCompound tagCompound, CallbackInfo ci) {
        final NBTTagList spawnlist = tagCompound.func_150295_c("Spawns", 10);
        for (int i = 0; i < spawnlist.func_74745_c(); i++) {
            final NBTTagCompound spawndata = spawnlist.func_150305_b(i);
            int spawndim = spawndata.func_74762_e("Dim");
            this.spawnChunkMap.put(spawndim, new BlockPos(spawndata.func_74762_e("SpawnX"), spawndata.func_74762_e("SpawnY"), spawndata.func_74762_e("SpawnZ")));
            this.spawnForcedMap.put(spawndim, spawndata.func_74767_n("SpawnForced"));
        }
    }

    @Inject(method = "writeEntityToNBT", at = @At(value = "FIELD", target = "net.minecraft.entity.player.EntityPlayer"
            + ".foodStats:Lnet/minecraft/util/FoodStats;"))
    public void onWriteEntityToNBT(NBTTagCompound tagCompound, CallbackInfo ci) {
        final NBTTagList spawnlist = new NBTTagList();
        for (Map.Entry<Integer, BlockPos> entry : this.spawnChunkMap.entrySet()) {
            BlockPos spawn = entry.getValue();
            if (spawn == null) {
                continue;
            }
            Boolean forced = this.spawnForcedMap.get(entry.getKey());
            if (forced == null) {
                forced = false;
            }
            NBTTagCompound spawndata = new NBTTagCompound();
            spawndata.func_74768_a("Dim", entry.getKey());
            spawndata.func_74768_a("SpawnX", spawn.func_177958_n());
            spawndata.func_74768_a("SpawnY", spawn.func_177956_o());
            spawndata.func_74768_a("SpawnZ", spawn.func_177952_p());
            spawndata.func_74757_a("SpawnForced", forced);
            spawnlist.func_74742_a(spawndata);
        }
        tagCompound.func_74782_a("Spawns", spawnlist);
    }

    public void setSpawnChunk(BlockPos pos, boolean forced, int dimension) {
        if (dimension == 0) {
            if (pos != null) {
                this.spawnChunk = pos;
                this.spawnForced = forced;
            } else {
                this.spawnChunk = null;
                this.spawnForced = false;
            }
            return;
        }

        if (pos != null) {
            this.spawnChunkMap.put(dimension, pos);
            this.spawnForcedMap.put(dimension, forced);
        } else {
            this.spawnChunkMap.remove(dimension);
            this.spawnForcedMap.remove(dimension);
        }
    }

    public BlockPos getBedLocation(int dimension) {
        return dimension == 0 ? this.spawnChunk : this.spawnChunkMap.get(dimension);
    }

    public boolean isSpawnForced(int dimension) {
        if (this.field_71093_bK == 0) {
            return this.spawnForced;
        }
        final Boolean forced = this.spawnForcedMap.get(dimension);
        return forced == null ? false : forced;
    }
}
